import math
import numpy as np
from . import sigproc
from . import feature_extractor


class Spectrogram(feature_extractor.FeaturesExtractor):

    def __init__(self,
                 features_num: int,
                 samplerate: int,
                 winlen: float,
                 winstep: float,
                 winfunc=None,
                 is_standardization=True,
                 pad_audio_to: int = 0):
        self.features_num = features_num
        self.winfunc = winfunc
        self.frame_len = int(winlen * samplerate)
        self.frame_step = int(winstep * samplerate)
        self.is_standardization = is_standardization
        self.pad_to = pad_audio_to

    def make_features(self, audio: np.ndarray) -> np.ndarray:
        """ Use `python_speech_features` lib to extract log-spectrogram's. """
        audio = self.normalize(audio.astype(np.float32))
        audio = (audio * np.iinfo(np.int16).max).astype(np.int16)
        audio = self.pad(audio) if self.pad_to else audio
        frames = sigproc.framesig(
            audio, self.frame_len, self.frame_step, self.winfunc
        )
        features = sigproc.logpowspec(
            frames, self.frame_len, norm=1
        )
        features = features[:, :self.features_num]  # Cut high frequency part
        return self.standardize(features) if self.is_standardization else features

    def pad(self, audio: np.ndarray) -> np.ndarray:
        """ Padding signal is required if you play with mixed precession. """
        length = 1 + int((len(audio) - self.frame_len) // self.frame_step + 1)
        pad_size = (self.pad_to - length % self.pad_to) * self.frame_step
        return np.pad(audio, (0, pad_size), mode='constant')



import tensorflow as tf
import os
from deepasr.utils import load_data
import deepasr as asr


def load(directory: str):
    """ Load each component of the CTC pipeline. """

    _network = tf.keras.models.load_model(os.path.join(directory, 'network.h5'))
    _network.load_weights(os.path.join(directory, 'model_weights.h5'))
    _optimizer = load_data(os.path.join(directory, 'optimizer.bin'))
    _alphabet = load_data(os.path.join(directory, 'alphabet.bin'))
    _decoder = load_data(os.path.join(directory, 'decoder.bin'))
    _features_extractor = load_data(
        os.path.join(directory, 'feature_extractor.bin'))
    _multi_gpu_flag = load_data(os.path.join(directory, 'multi_gpu_flag.bin'))
    _sample_rate = load_data(os.path.join(directory, 'sample_rate.bin'))

    pipeline = asr.pipeline.ctc_pipeline.CTCPipeline(
        alphabet=_alphabet, features_extractor=_features_extractor, model=_network, optimizer=_optimizer,
        decoder=_decoder, sample_rate=_sample_rate, mono=True, multi_gpu=_multi_gpu_flag
    )
    return pipeline



import abc
from typing import List
import numpy as np
import pandas as pd
from tensorflow import keras
import sys

sys.path.append("..")
from deepasr.decoder import Decoder
from deepasr.features import FeaturesExtractor
from deepasr.vocab import Alphabet


class Pipeline:

    @property
    @abc.abstractmethod
    def alphabet(self) -> Alphabet:
        pass

    @property
    @abc.abstractmethod
    def features_extractor(self) -> FeaturesExtractor:
        pass

    @property
    @abc.abstractmethod
    def model(self) -> keras.Model:
        pass

    @property
    @abc.abstractmethod
    def decoder(self) -> Decoder:
        pass

    @abc.abstractmethod
    def fit(self,
            train_dataset: pd.DataFrame,
            val_dataset: pd.DataFrame,
            prepared_features=False,
            **kwargs) -> keras.callbacks.History:
        pass

    @abc.abstractmethod
    def predict(self, batch_audio: List[np.ndarray], **kwargs) -> List[str]:
        pass

    @abc.abstractmethod
    def save(self, directory: str):
        pass



import abc
import itertools
from typing import List
import numpy as np


class Decoder:

    @abc.abstractmethod
    def __call__(self, batch_logits: np.ndarray) -> List[np.ndarray]:
        pass


class GreedyDecoder:

    def __call__(self, batch_logits: np.ndarray) -> List[np.ndarray]:
        """ Decode the best guess from logits using greedy algorithm. """
        # Choose the class with maximum probability
        best_candidates = np.argmax(batch_logits, axis=2)
        # Merge repeated chars
        decoded = [np.array([k for k, _ in itertools.groupby(best_candidate)])
                   for best_candidate in best_candidates]
        return decoded



from typing import Tuple
import numpy as np


class SpecAugment:

    def __init__(self,
                 F: int = None,
                 mf: int = None,
                 Tmin: int = None,
                 Tmax: int = None,
                 mt: int = None):
        """ SpecAugment: A Simple Data Augmentation Method. """
        self.F = F
        self.mf = mf
        self.Tmin = Tmin
        self.Tmax = Tmax
        self.mt = mt

    def __call__(self, batch_features: np.ndarray) -> np.ndarray:
        return np.stack([self.mask_features(features) for features in batch_features], axis=0)

    def mask_features(self, features: np.ndarray) -> np.ndarray:
        features = features.copy()
        time, channels = features.shape
        means = features.mean(axis=0)  # The mean should be zero if features are normalized
        if self.F and self.mf:
            features = self.mask_frequencies(features, means, channels, self.F, self.mf)
        if self.Tmax and self.mt:
            features = self.mask_time(features, means, time, (self.Tmin, self.Tmax), self.mt)
        return features

    @staticmethod
    def mask_frequencies(features: np.ndarray, means: np.ndarray, channels: int, F: int, mf: int):
        for i in range(mf):
            f = np.random.random_integers(low=0, high=F)
            f0 = np.random.random_integers(low=0, high=channels - F)
            features[:, f0:f0 + f] = means[f0:f0 + f]
        return features

    @staticmethod
    def mask_time(features: np.ndarray, means: np.ndarray, time: int, T_range: Tuple[int, int], mt: int):
        Tmin, Tmax = T_range
        for i in range(mt):
            t = np.random.random_integers(low=Tmin, high=Tmax)
            t0 = np.random.random_integers(low=0, high=time - Tmax)
            features[t0:t0 + t, :] = means
        return features



import itertools
from typing import Tuple, List
from collections import defaultdict
import numpy as np


def edit_distance(source: List[str],
                  destination: List[str]) -> Tuple[int, np.ndarray, np.ndarray]:
    """
    Calculation of edit distance between two sequences.

    This is the Levenshtein distance with the substitution cost equals 1.
    It is the iterative method with the full matrix support.
    O(nm) time and space complexity.

    References:
    https://web.stanford.edu/class/cs124/lec/med.pdf
    https://www.python-course.eu/levenshtein_distance.php
    """
    size_x = len(source) + 1
    size_y = len(destination) + 1
    matrix = np.zeros([size_x, size_y])
    matrix[:, 0] = np.arange(0, size_x)
    matrix[0, :] = np.arange(0, size_y)
    backtrace = np.zeros_like(matrix, dtype=[('del', bool),
                                             ('sub', bool),
                                             ('ins', bool),
                                             ('cost', int)])
    backtrace[:, 0] = (True, False, False, 0)
    backtrace[0, :] = (False, False, True, 0)
    for x, y in itertools.product(range(1, size_x),
                                  range(1, size_y)):
        if source[x - 1] == destination[y - 1]:
            cost = 0
        else:
            cost = 1
        delete = matrix[x - 1][y] + 1
        insert = matrix[x][y - 1] + 1
        substitute = matrix[x - 1][y - 1] + cost
        min_dist = min(delete, insert, substitute)
        matrix[x, y] = min_dist
        backtrace[x, y] = (delete == min_dist,
                           substitute == min_dist,
                           insert == min_dist,
                           cost)
    return matrix[size_x - 1, size_y - 1], matrix, backtrace


def simple_backtrace(backtrace: np.ndarray):
    """ Calculate the editing path via the backtrace. """
    rows, columns = backtrace.shape
    i, j = rows - 1, columns - 1
    backtrace_indices = [(i, j, 'sub', 0)]
    while (i, j) != (0, 0):
        delete, substitute, insert, cost = backtrace[i, j]
        if insert:
            operation = 'ins'
            i, j = i, j - 1
        elif substitute:
            operation = 'sub'
            i, j = i - 1, j - 1
        elif delete:
            operation = 'del'
            i, j = i - 1, j
        else:
            raise KeyError("Backtrace matrix wrong defined")
        backtrace_indices.append((i, j, operation, cost))
    return list(reversed(backtrace_indices))


def decode_path(best_path: List[Tuple[int, int, str, int]],
                source: List[str],
                destination: List[str]):
    """ Collect all transformations needed to go from `source` to
    `destination`. """
    to_delete, to_insert, to_substitute = [], [], defaultdict(list)
    for index, (i, j, operation, cost) in enumerate(best_path):
        if operation == 'del':
            item = source[i]
            to_delete.append(item)
        elif operation == 'sub' and cost:
            # without cost sub operation indicates correctness
            wrong_item, target_item = source[i], destination[j]
            to_substitute[target_item].append(wrong_item)
        elif operation == 'ins':
            item = destination[j]
            to_insert.append(item)
    return to_delete, to_insert, to_substitute



import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model
from tensorflow.keras.layers import *
from tensorflow.keras.mixed_precision import experimental as mixed_precision


def get_deepasrnetwork1(input_dim=None, output_dim=29,
                        is_mixed_precision=True, random_state=1) -> keras.Model:
    """

    input_dim: int i wielokrotność 4
    output_dim: licba liter w słowniku

    """
    if is_mixed_precision:
        policy = mixed_precision.Policy('float32')
        mixed_precision.set_policy(policy)

    np.random.seed(random_state)
    tf.random.set_seed(random_state)

    # the input
    input_data = Input(name='the_input', shape=(None, input_dim), dtype='float32')

    # Batch normalize
    bn1 = BatchNormalization(axis=-1, name='BN_1')(input_data)

    # 1D Convs
    conv = Conv1D(filters=220, kernel_size=5, strides=2, padding='valid', activation='relu', name='Conv1D_1')(bn1)
    conv = BatchNormalization(name="CNBN_1")(conv)
    conv1 = Conv1D(filters=220, kernel_size=5, strides=2, padding='valid', activation='relu', name='Conv1D_2')(conv)
    conv1 = BatchNormalization(name="CNBN_2")(conv1)

    # RNN
    gru_1 = GRU(512, return_sequences=True, name='gru_1')(conv1)
    gru_2 = GRU(512, return_sequences=True, go_backwards=True, name='gru_2')(conv1)

    # merge tow gpu ouputs
    merged = concatenate([gru_1, gru_2])
    # Batch normalize
    bn2 = BatchNormalization(axis=-1, name="BN_2")(merged)

    dense = TimeDistributed(Dense(30))(bn2)
    y_pred = TimeDistributed(Activation('softmax', name='softmax'))(dense)

    model = Model(inputs=input_data, outputs=y_pred)

    return model
